/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements. See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership. The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
*/

package org.apache.yoko.tools.processors.idl;

import javax.wsdl.Definition;

import antlr.collections.AST;

import org.apache.schemas.yoko.bindings.corba.ModeType;
import org.apache.schemas.yoko.bindings.corba.OperationType;
import org.apache.schemas.yoko.bindings.corba.ParamType;
import org.apache.schemas.yoko.bindings.corba.TypeMappingType;

import org.apache.ws.commons.schema.XmlSchema;
import org.apache.ws.commons.schema.XmlSchemaCollection;
import org.apache.ws.commons.schema.XmlSchemaElement;
import org.apache.ws.commons.schema.XmlSchemaSequence;
import org.apache.ws.commons.schema.XmlSchemaType;

import org.apache.yoko.wsdl.CorbaTypeImpl;

public class ParamDclVisitor extends VisitorBase {

    private XmlSchemaSequence inWrappingSequence; 
    private XmlSchemaSequence outWrappingSequence;
    private OperationType corbaOperation;
    private Definition wsdlDefinition;
    
    public ParamDclVisitor(Scope scope,
                           XmlSchemaCollection xmlSchemas,
                           XmlSchema xmlSchema,
                           TypeMappingType corbaTypeMap,
                           XmlSchemaSequence inWrapSeq,
                           XmlSchemaSequence outWrapSeq,
                           OperationType corbaOp) {
        super(scope, xmlSchemas, xmlSchema, corbaTypeMap);
        inWrappingSequence = inWrapSeq;
        outWrappingSequence = outWrapSeq;
        corbaOperation = corbaOp;
    }

    public static boolean accept(AST node) {
        boolean result = false;
        if (node != null) {
            int type = node.getType();
            result = type == IDLTokenTypes.LITERAL_in
                || type == IDLTokenTypes.LITERAL_out
                || type == IDLTokenTypes.LITERAL_inout;
        }
        return result;
    }
    
    public void visit(AST node) {
        // <param_dcl> ::= <param_attribute> <param_type_spec> <simple_declarator>
        // <param_attribute> ::= "in"
        //                     | "out"
        //                     | "inout"

        AST typeNode = node.getFirstChild();
        AST nameNode = TypesUtils.getCorbaTypeNameNode(typeNode);
        
        ParamTypeSpecVisitor visitor = new ParamTypeSpecVisitor(getScope(), 
                                                                schemas, 
                                                                schema, 
                                                                typeMap,
                                                                wsdlDefinition);
        visitor.visit(typeNode);
        XmlSchemaType schemaType = visitor.getSchemaType();
        CorbaTypeImpl corbaType = visitor.getCorbaType();
        
        boolean isObjectReference = false;
        if (corbaType instanceof org.apache.schemas.yoko.bindings.corba.Object) {
            isObjectReference = true;
        }
        
        switch (node.getType()) {
        case IDLTokenTypes.LITERAL_in: 
            addElement(inWrappingSequence, schemaType, nameNode.toString(), isObjectReference);
            addCorbaParam(corbaType, ModeType.IN, nameNode.toString());
            break;
        case IDLTokenTypes.LITERAL_out:
            addElement(outWrappingSequence, schemaType, nameNode.toString(), isObjectReference);            
            addCorbaParam(corbaType, ModeType.OUT, nameNode.toString());
            break;
        case IDLTokenTypes.LITERAL_inout:
            addElement(inWrappingSequence, schemaType, nameNode.toString(), isObjectReference);
            addElement(outWrappingSequence, schemaType, nameNode.toString(), isObjectReference);
            addCorbaParam(corbaType, ModeType.INOUT, nameNode.toString());
            break;
        default:
            throw new RuntimeException("[ParamDclVisitor: illegal IDL!]");
        }
            
        
        setSchemaType(schemaType);
        setCorbaType(corbaType);
    }

    // need this to pass to our parameter type spec in the case of an 
    // object refernce.
    public void setWSDLDefinition(Definition wsdlDef) {
        wsdlDefinition = wsdlDef;
    }
    
    private XmlSchemaElement addElement(XmlSchemaSequence schemaSequence,
                                        XmlSchemaType schemaType,
                                        String name,
                                        boolean isObjectReference) {
        XmlSchemaElement element = new XmlSchemaElement();
        element.setName(name);
        // REVISIT: An object reference requires that we use a ref name.  An element type produces
        // a wsdltojava error.
        if (isObjectReference) {
            element.setRefName(schemaType.getQName());
        } else {
            element.setSchemaTypeName(schemaType.getQName());
        }
        schemaSequence.getItems().add(element);
        return element;
    }

    private void addCorbaParam(CorbaTypeImpl corbaType, ModeType mode, String partName) {
        ParamType param = new ParamType();
        param.setName(partName);
        param.setMode(mode);
        param.setIdltype(corbaType.getQName());
        corbaOperation.getParam().add(param);
    }

}
