/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements. See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership. The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
*/

package org.apache.yoko.tools.processors.idl;

import java.util.Iterator;

import javax.xml.namespace.QName;

import antlr.collections.AST;

import org.apache.schemas.yoko.bindings.corba.TypeMappingType;

import org.apache.ws.commons.schema.XmlSchema;
import org.apache.ws.commons.schema.XmlSchemaCollection;
import org.apache.ws.commons.schema.XmlSchemaType;

import org.apache.ws.commons.schema.constants.Constants;

import org.apache.yoko.wsdl.CorbaConstants;
import org.apache.yoko.wsdl.CorbaTypeImpl;

public class ScopedNameVisitor extends VisitorBase {

    public ScopedNameVisitor(Scope scope,
                             XmlSchemaCollection xmlSchemas,
                             XmlSchema xmlSchema,
                             TypeMappingType corbaTypeMap) {
        super(scope, xmlSchemas, xmlSchema, corbaTypeMap);
    }
    
    public static boolean accept(Scope scope,
                                 XmlSchemaCollection schemas,
                                 XmlSchema schema,
                                 TypeMappingType typeMap,                                 
                                 AST node) {
        boolean result = false;
        if (PrimitiveTypesVisitor.accept(node)) {
            result = true;
        } else if (findSchemaType(scope, schemas, schema, typeMap, node) != null) {
            result = true;
        }
        return result;
    }

    public void visit(AST node) {
        // <scoped_name> ::= <identifier>
        //                 | :: <identifier>
        //                 | <scoped_name> "::" <identifier>

        XmlSchemaType stype = null;
        CorbaTypeImpl ctype = null;
        
        if (PrimitiveTypesVisitor.accept(node)) {
            // primitive type
            
            PrimitiveTypesVisitor primitiveVisitor = new PrimitiveTypesVisitor(null, schemas);
            primitiveVisitor.visit(node);
            
            stype = primitiveVisitor.getSchemaType();
            ctype = primitiveVisitor.getCorbaType();
        } else {
            // declared type
            Scope currentScope = getScope();
            while (stype == null
                && currentScope != currentScope.getParent()) {
                // A name can be used in an unqualified form within a particular scope;
                // it will be resolved by successvely searching farther out in enclosing
                // scopes, while taking into consideration inheritance relationships 
                // among interfaces. 
                // INHERITANCE NOT IMPLEMENTED YET
                Scope scopedName = new Scope(currentScope, node);
                QName schemaQName = new QName(schema.getTargetNamespace(), scopedName.toString());

                stype = schema.getTypeByName(schemaQName);

                currentScope = currentScope.getParent();
            }
            if (stype == null) {
                // Global scope is our last chance to resolve the node
                QName schemaQName = new QName(schema.getTargetNamespace(), node.toString());
                stype = schema.getTypeByName(schemaQName);
            }
                
            // handle special case of simple string alias
            // a plain xsd:string has no corresponding XmlSchemaType in the XmlSchema
            // so the previous findType() method would return null.
            // As a temporary workaround, we query the typeMap for a corba:string with
            // same name. If the string is there, then it had been previously properly
            // declared and we can use it here.
            if (stype == null) {
                QName corbaQName = null;
                currentScope = getScope();
                while (ctype == null
                    && currentScope != currentScope.getParent()) {
                    Scope scopedName = new Scope(currentScope, node);
                    corbaQName = new QName(typeMap.getTargetNamespace(), scopedName.toString());
                    ctype = findCorbaType(typeMap, corbaQName);
                    
                    currentScope = currentScope.getParent();
                }
                if (ctype == null) {
                    corbaQName = new QName(typeMap.getTargetNamespace(), node.toString());
                    ctype = findCorbaType(typeMap, corbaQName);                    
                }
                if (ctype != null && ctype.getType() == Constants.XSD_STRING) {
                    stype = schemas.getTypeByQName(Constants.XSD_STRING);
                    ctype = new CorbaTypeImpl();
                    ctype.setName(CorbaConstants.NT_CORBA_STRING.getLocalPart());
                    ctype.setQName(CorbaConstants.NT_CORBA_STRING);
                    ctype.setType(Constants.XSD_STRING);
                } else {
                    throw new RuntimeException("[ScopedNameVisitor: Corba type "
                                               + corbaQName.toString()
                                               + " not found in typeMap]");
                }
            } else {
                ctype = findCorbaType(typeMap, stype.getQName());
            }
        }

        setSchemaType(stype);
        setCorbaType(ctype);
        
    }
    
    private static XmlSchemaType findSchemaType(Scope scope,
                                                XmlSchemaCollection schemas,
                                                XmlSchema schema,
                                                TypeMappingType typeMap,
                                                AST node) {
        XmlSchemaType result = null; 
        Scope currentScope = scope;
        while (result == null
            && currentScope != currentScope.getParent()) {
            // A name can be used in an unqualified form within a particular scope;
            // it will be resolved by successvely searching farther out in enclosing
            // scopes, while taking into consideration inheritance relationships 
            // among interfaces. 
            // INHERITANCE NOT IMPLEMENTED YET
            Scope scopedName = new Scope(currentScope, node);
            QName qname = new QName(schema.getTargetNamespace(), scopedName.toString());
            result = schema.getTypeByName(qname);
            currentScope = currentScope.getParent();
        }
        if (result == null) {
            // Global scope is our last chance to resolve the node
            QName qname = new QName(schema.getTargetNamespace(), node.toString());
            result = schema.getTypeByName(qname);
        }
            
        // handle special case of simple string alias
        // a plain xsd:string has no corresponding XmlSchemaType in the XmlSchema
        // so the previous findType() method would return null.
        // As a temporary workaround, we query the typeMap for a corba:string with
        // same name. If the string is there, then it had been previously properly
        // declared and we can use it here.
        if (result == null) {
            CorbaTypeImpl ctype = null;
            currentScope = scope;
            while (ctype == null
                && currentScope != currentScope.getParent()) {
                QName corbaStringQname = new QName(typeMap.getTargetNamespace(), currentScope.toString());
                ctype = findCorbaType(typeMap, corbaStringQname);
                currentScope = currentScope.getParent();
            }
            if (ctype == null) {
                QName corbaStringQname = new QName(typeMap.getTargetNamespace(), node.toString());
                ctype = findCorbaType(typeMap, corbaStringQname);                
            }
            if (ctype != null
                && ctype.getType() == Constants.XSD_STRING) {
                result = schemas.getTypeByQName(Constants.XSD_STRING);
            }
        }
        return result;
    }
    
    public static CorbaTypeImpl findCorbaType(TypeMappingType typeMap, QName schemaTypeName) {
        CorbaTypeImpl result = null;
        Iterator corbaTypes = typeMap.getStructOrExceptionOrUnion().iterator();
        while (corbaTypes.hasNext()) {
            CorbaTypeImpl type = (CorbaTypeImpl) corbaTypes.next();
            if (type.getQName().getLocalPart().equals(schemaTypeName.getLocalPart())) {
                result = type;
                break;
            }
        }
        return result;
    }

}
