/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements. See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership. The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
*/

package org.apache.yoko.tools.processors.idl;

import javax.xml.namespace.QName;

import antlr.collections.AST;

import org.apache.schemas.yoko.bindings.corba.CaseType;
import org.apache.schemas.yoko.bindings.corba.TypeMappingType;
import org.apache.schemas.yoko.bindings.corba.Union;
import org.apache.schemas.yoko.bindings.corba.Unionbranch;

import org.apache.ws.commons.schema.XmlSchema;
import org.apache.ws.commons.schema.XmlSchemaChoice;
import org.apache.ws.commons.schema.XmlSchemaCollection;
import org.apache.ws.commons.schema.XmlSchemaComplexType;
import org.apache.ws.commons.schema.XmlSchemaElement;
import org.apache.ws.commons.schema.XmlSchemaSequence;
import org.apache.ws.commons.schema.XmlSchemaType;

import org.apache.yoko.wsdl.CorbaTypeImpl;

public class UnionVisitor extends VisitorBase {

    private final String discriminator = "discriminator";
    
    public UnionVisitor(Scope scope,
                        XmlSchemaCollection xmlSchemas,
                        XmlSchema xmlSchema,
                        TypeMappingType typeMappingType) {
        super(scope, xmlSchemas, xmlSchema, typeMappingType);
    }
    
    public static boolean accept(AST node) {
        if (node.getType() == IDLTokenTypes.LITERAL_union) {
            return true;
        }
        return false;
    }
    
    public void visit(AST unionNode) {
        // <union_type> ::= "union" <identifier> "switch" "(" <switch_type_spec> ")"
        //                  "{" <switch_body> "}"
        // <switch_type_spec> ::= <integer_type>
        //                      | <char_type>
        //                      | <boolean_type>
        //                      | <enum_type>
        //                      | <scoped_type>
        // <switch_body> ::= <case>+
        // <case> ::= <case_label>+ <element_spec> ";"
        // <case_label> ::= "case" <const_expr> ":"
        //                | "default" ":"
        // <element_spec> ::= <type_spec> <declarator>
        
        
        AST identifierNode = unionNode.getFirstChild();
        Scope unionScope = new Scope(getScope(), identifierNode);
        AST discriminatorNode = identifierNode.getNextSibling();
        AST caseNode = discriminatorNode.getNextSibling();

        // xmlschema:union
        XmlSchemaComplexType unionSchemaComplexType = new XmlSchemaComplexType(schema);
        XmlSchemaSequence sequence = new XmlSchemaSequence();
        unionSchemaComplexType.setName(unionScope.toString());
        unionSchemaComplexType.setParticle(sequence);
        
        // REVISIT
        // TEMPORARILY 
        // using TypesVisitor to visit <const_type>
        // it should be visited by a SwitchTypeSpecVisitor
        TypesVisitor visitor = new TypesVisitor(getScope(), schemas, schema, typeMap, null);
        visitor.visit(discriminatorNode);
        XmlSchemaType stype = visitor.getSchemaType();
        CorbaTypeImpl ctype = visitor.getCorbaType();
        
        XmlSchemaElement discriminatorElement = new XmlSchemaElement();
        discriminatorElement.setName(discriminator);
        discriminatorElement.setSchemaTypeName(stype.getQName());
        discriminatorElement.setMinOccurs(1);
        discriminatorElement.setMaxOccurs(1);
        sequence.getItems().add(discriminatorElement);
        
        XmlSchemaChoice choice = new XmlSchemaChoice();
        choice.setMinOccurs(1);
        choice.setMaxOccurs(1);
        sequence.getItems().add(choice);
        
        
        // corba:union
        Union corbaUnion = new Union();
        corbaUnion.setQName(new QName(typeMap.getTargetNamespace(), unionScope.toString()));
        corbaUnion.setRepositoryID(unionScope.toIDLRepositoryID());
        corbaUnion.setType(unionSchemaComplexType.getQName());
        corbaUnion.setDiscriminator(ctype.getQName());
        
        processCaseNodes(caseNode, unionScope, choice, corbaUnion);

        // add schemaType
        schema.getItems().add(unionSchemaComplexType);
        schema.addType(unionSchemaComplexType);

        // add corbaType
        typeMap.getStructOrExceptionOrUnion().add(corbaUnion);

        // REVISIT: are these assignments needed?
        setSchemaType(unionSchemaComplexType);
        setCorbaType(corbaUnion);
    }
    
    private void processCaseNodes(AST caseNode,
                                  Scope scope,
                                  XmlSchemaChoice choice,
                                  Union corbaUnion) {
        while (caseNode != null) {
            AST typeNode  = null;
            AST nameNode  = null;
            AST labelNode = null;
            
            // xmlschema:element
            XmlSchemaElement element = new XmlSchemaElement();

            // corba:unionbranch
            Unionbranch unionBranch = new Unionbranch();

            if (caseNode.getType() == IDLTokenTypes.LITERAL_default) {
                // default:
                unionBranch.setDefault(true);
                
                typeNode = caseNode.getFirstChild();
                nameNode = typeNode.getNextSibling();
            } else {
                // case:
                createCase(caseNode, unionBranch);
                
                labelNode = caseNode.getFirstChild();
                if (labelNode.getType() == IDLTokenTypes.LITERAL_case) {
                    labelNode = labelNode.getNextSibling();
                }
                
                typeNode = labelNode.getNextSibling();
                nameNode = typeNode.getNextSibling();
            }
            

            TypesVisitor visitor = new TypesVisitor(scope,
                                                    schemas,
                                                    schema,
                                                    typeMap,
                                                    null);
            visitor.visit(typeNode);
            XmlSchemaType stype = visitor.getSchemaType();
            CorbaTypeImpl ctype = visitor.getCorbaType();
            
            
            // needed for anonymous arrays in unions
            if (ArrayVisitor.accept(nameNode)) {
                Scope anonScope = new Scope(scope, TypesUtils.getCorbaTypeNameNode(nameNode));
                ArrayVisitor arrayVisitor = new ArrayVisitor(anonScope,
                                                             schemas,
                                                             schema,
                                                             typeMap,
                                                             stype,
                                                             ctype,
                                                             null);
                arrayVisitor.visit(nameNode);
                stype = arrayVisitor.getSchemaType();
                ctype = arrayVisitor.getCorbaType();
            }
            
            
            // xmlschema:element
            element.setName(nameNode.toString());
            element.setSchemaTypeName(stype.getQName());
            choice.getItems().add(element);
            
            
            // corba:unionbranch
            unionBranch.setName(nameNode.toString());
            unionBranch.setIdltype(ctype.getQName());
            corbaUnion.getUnionbranch().add(unionBranch);
            
            
            caseNode = caseNode.getNextSibling();
        }
    }
    
    private void createCase(AST caseNode, Unionbranch unionBranch) {
        AST node = caseNode.getFirstChild();
        if (node != null) {
            if (node.getType() == IDLTokenTypes.LITERAL_case) {
                // corba:case
                CaseType caseType = new CaseType();
                caseType.setLabel(node.getNextSibling().toString());
                unionBranch.getCase().add(caseType);
                
                // recursive call
                createCase(node, unionBranch);
            } else {
                // corba:case
                CaseType caseType = new CaseType();
                caseType.setLabel(node.toString());
                unionBranch.getCase().add(caseType);
            }
        }
    }
}
